import 'dart:collection';

import 'package:text_table/text_table.dart';
import 'package:grizzly_range/grizzly_range.dart' as librange;

extension NullableIterableExt<T> on Iterable<T?> {
  int count(T? v) {
    int ret = 0;

    for (final item in this) {
      if (item == v) {
        ret++;
      }
    }

    return ret;
  }

  int get countNonNull {
    int ret = 0;
    for (final item in this) {
      if (item != null) ret++;
    }
    return ret;
  }

  Iterable<T?> selectByMask(Iterable<bool> mask) sync* {
    Iterator<T?> iter1 = iterator;
    Iterator<bool> iter2 = mask.iterator;

    while (true) {
      final has1 = iter1.moveNext();
      final has2 = iter2.moveNext();

      if (!has1 || !has2) {
        break;
      }

      if (iter2.current) {
        yield iter1.current;
      }
    }
  }

  Iterable<int> get indices => librange.indices(length);

  bool isEqual(Iterable<T?> v) {
    if (length != v.length) return false;
    final iter1 = iterator;
    final iter2 = v.iterator;
    while (iter1.moveNext() && iter2.moveNext()) {
      if (iter1.current != iter2.current) {
        return false;
      }
    }
    return true;
  }

  bool isEqualScalar(T v) {
    for (final item in this) {
      if (v != item) return false;
    }
    return true;
  }
}

extension IterableExt<T> on Iterable<T> {
  T operator [](int index) => elementAt(index);

  Iterable<T> cut(int start, [int? end]) {
    Iterable<T> ret = skip(start);
    if (end != null) {
      if (end <= 0) {
        end = length - 1 + end;
        ret = ret.take(end - start + 1);
      } else {
        ret = ret.take(end - start + 1);
      }
    }
    return ret;
  }

  List<T> unique() {
    // ignore: prefer_collection_literals
    final ret = LinkedHashSet<T>();
    for (T v in this) {
      if (!ret.contains(v)) ret.add(v);
    }
    return ret.toList();
  }

  Map<T, int> valueCounts() {
    final groups = <T, int>{};

    for (final item in this) {
      groups[item] = (groups[item] ?? 0) + 1;
    }

    return groups;
  }

  Iterable<T> selectByMask(Iterable<bool> mask) sync* {
    Iterator<T> iter1 = iterator;
    Iterator<bool> iter2 = mask.iterator;

    while (true) {
      final has1 = iter1.moveNext();
      final has2 = iter2.moveNext();

      if (!has1 || !has2) {
        break;
      }

      if (iter2.current) {
        yield iter1.current;
      }
    }
  }

  Iterable<S> same<S>(S v) => map((e) => v);

  List<List<T>> toRow() => [toList()];

  List<List<T>> toCol() => map((e) => [e]).toList();

  Iterable<Iterable<T>> diagonal(T fill) => Iterable.generate(length,
      (r) => Iterable.generate(length, (c) => r == c ? elementAt(r) : fill));

  Iterable<String> toStringView() => map((e) => (e ?? '').toString());

  String tableString({TableRenderer renderer = const TableRenderer()}) {
    return renderer.render(toStringView().toCol(),
        columns: librange.indices(length).map((i) => i.toString()).toList());
  }
}
